--[[ Kniggel ]]

--TODOs:

local iTP = LibStub:GetLibrary("iTransferProtocol-1.0")
if not iTP then return end
local iTPCallback = iTP:RegisterPrefix("KNiGGEL")

local DefaultO = {
  ["framePoint"] = "CENTER";
  ["frameRelativeTo"] = "UIParent";
  ["frameRelativePoint"] = "CENTER";
  ["frameOffsetX"] = 0;
  ["frameOffsetY"] = 0;
  ["framePointPopup"] = "CENTER";
  ["frameRelativeToPopup"] = "UIParent";
  ["frameRelativePointPopup"] = "CENTER";
  ["frameOffsetXPopup"] = 0;
  ["frameOffsetYPopup"] = 0;
  ["showFrame"] = 1;
  ["invitelist"] = "";
  ["fixed"] = 0;
}
local O

local frame = CreateFrame("Frame", "KniggelFrame", UIParent)
local popupFrame = CreateFrame("Frame", "KniggelPopupFrame", UIParent)
local frameEvents = {};
local PLAYER_WIDTH = 100
local PLAYER_HEIGHT = 12
local DICE_WIDTH = 20
local ROLL_WIDTH = 40
local ROW_COUNT = 20
local COL_SPACING = 5
local isdebug = false
local function logdebug(...)
  if isdebug then
    print(...)
  end
end
local isInEncounter = false
local players = {};
local diceLock = {
  {1, 0},
  {1, 0},
  {1, 0},
  {1, 0},
  {1, 0},
};
--KNIGGELDICELOCK = diceLock
local myState = -1
local nextTurnPlayer = 0
local turnPlayerId = ""
local myGames = {};
local myGameId = ""

local invitedBy, invitedByGameId = "", ""
local myPlayerId

--forward function declarations
local updatePlayerColumnContent, mysplit2, updateGameState

local function addPlayerToGameResponded(gameId, playerId)
  if not myGames[gameId] then
    return
  end
  
  for i, v in ipairs(myGames[gameId].respondedlist) do
    if v == playerId then
      return
    end
  end
  table.insert(myGames[gameId].respondedlist, playerId)
end

local function addPlayerToGame(gameId, playerId)
  if not myGames[gameId] then
    return
  end
  
  for i, v in ipairs(myGames[gameId].playerlist) do
    if v == playerId then
      return
    end
  end
  table.insert(myGames[gameId].playerlist, playerId)
end

local function isPlayerInvited(gameId, playerId)
  if not myGames[gameId] then
    return false
  end
  
  for i, v in ipairs(myGames[gameId].playerlist) do
    if v == playerId then
      return true
    end
  end
  return false
end

--local function getNewPlayerTable(playername, realmname)
local function getNewPlayerTable(playerId)
  local i = playerId:find("-")
  local playername, realmname = playerId:sub(1, i - 1), playerId:sub(i + 1)
  return {
    ["id"] = format("%s-%s", playername, realmname);
    ["nameshort"] = playername;
    ["field"] = {
      [1] = -1;
      [2] = -1;
      [3] = -1;
      [4] = -1;
      [5] = -1;
      [6] = -1;
      [7] = -1; --3 of a kind                              (sum)
      [8] = -1; --4 of a kind                              (sum)
      [9] = -1; --full house (3 of a kind + 2 of a kind)  25
      [10] = -1; --small street (sequence of 4)            30
      [11] = -1; --large street (sequence of 5)            40
      [12] = -1; --5 of a kind                             50
      [13] = -1; --chance                                  (sum)
    };
    ["collink"] = nil; --link to column table
  };
end

--local function getPlayerTableById(playername, realmname)
local function getPlayerTableById(playerId)
  --local id = format("%s-%s", playername, realmname)
  for i, t in ipairs(players) do
    --if t.id == id then
    if t.id == playerId then
      return i, t
    end
  end
  return 0, nil
end

--local function addPlayerTable(playername, realmname)
local function addPlayerTable(playerId)
  --local i = getPlayerTableById(playername, realmname)
  local i = getPlayerTableById(playerId)
  if i > 0 then
    return
  end
  
  --local t = getNewPlayerTable(playername, realmname)
  local t = getNewPlayerTable(playerId)
  table.insert(players, t)
  
  frame:addPlayerColumn(t)
  frame:alignPlayerColumns()
  updatePlayerColumnContent(t)
  
  return t
end

local function removePlayerTableById(playerId)--(playername, realmname)
  --local i, t = getPlayerTableById(playername, realmname)
  local i, t = getPlayerTableById(playerId)
  if i == 0 then
    return
  end
  
  frame:removePlayerColumn(t)
  players[i] = nil
end

local function resetPlayer(playerTable)
  for i = 1, 13 do
    playerTable["field"][i] = -1
  end
end

local function removeAllPlayerTables()
  while #players > 0 do
    removePlayerTableById(players[1].id)
  end
  
  --invitedPlayers = {};
end

local function addGame(gameId, invitedBy)
  local game = {
    ["invitedBy"] = invitedBy;
    ["playerlist"] = {};
    ["respondedlist"] = {};
    ["state"] = 0;
  };
  myGames[gameId] = game
  return game
end

--playerIdList = "player1-realm,player2-realm,..."
local function inviteNewGame(playerIdList)
  myGameId = frame:getNewGameId()
  removeAllPlayerTables()
  addPlayerTable(myPlayerId)
  if not playerIdList then
    local game = addGame(myGameId, myPlayerId)
    --TODO: fill game.playerlist with players in RAID/GUILD (active channel[s])
    frame:sendBroadcastInvite(myGameId)
  else
    local game = addGame(myGameId, myPlayerId)
    game.playerlist = mysplit2(playerIdList, ",")
    for i, v in ipairs(game.playerlist) do
      frame:sendPersonalInvite(myGameId, v)
    end
    addPlayerToGame(myGameId, myPlayerId) --don't need an invite, so add now
    addPlayerToGameResponded(myGameId, myPlayerId) --i responded to my own game
  end
  
  myState = 0
  updateGameState()
  
  --TODO
  --...
end

local function buildGameField(gameId)
  removeAllPlayerTables()
  local game = myGames[gameId]
  for i, v in ipairs(game.playerlist) do
    addPlayerTable(v)
  end
end

local function prepareState(state)
  if state == 1 then
    --choose next player
    nextTurnPlayer = nextTurnPlayer + 1
    if nextTurnPlayer > #(myGames[myGameId].playerlist) then
      nextTurnPlayer = 1
    end
    turnPlayerId = myGames[myGameId].playerlist[nextTurnPlayer]
    logdebug("turnPlayerId", turnPlayerId)
    
    frame:sendBroadcastPlayersTurn(myGameId, 1, turnPlayerId)
  elseif state == 2 then
    frame:sendBroadcastPlayersTurn(myGameId, 2, turnPlayerId)
  elseif state == 3 then
    frame:sendBroadcastPlayersTurn(myGameId, 3, turnPlayerId)
  elseif state == 4 then
    frame:sendBroadcastPlayersTurn(myGameId, 4, turnPlayerId)
  elseif state == 5 then
    frame:sendBroadcastPlayersTurn(myGameId, 5, turnPlayerId)
  elseif state == 6 then
    frame:sendBroadcastPlayersTurn(myGameId, 6, turnPlayerId)
  elseif state == 7 then
    frame:sendBroadcastPlayersTurn(myGameId, 7, turnPlayerId)
  elseif state == 8 then
    local winPlayerId, winPlayerPoints = "", 0
    for _, playerId in ipairs(myGames[myGameId].playerlist) do
      local _, t = getPlayerTableById(playerId)
      local points = updatePlayerColumnContent(t)
      if points > winPlayerPoints then
        winPlayerPoints = points
        winPlayerId = playerId
      end
    end
    frame:sendBroadcastPlayersTurn(myGameId, 8, winPlayerId) --turnPlayerId = winner playerId
  end
end

local function rollDice()
  for i = 1, 5 do
    if diceLock[i][2] == 0 then
      diceLock[i][1] = random(1, 6)
    end
  end
  frame:updateDiceDisplay()
  frame:sendMyDiceStatus(myGameId)
end

--call lockDice([01], [01], [01], [01], [01])
local function lockDice(...)
  for i = 1, 5 do
    if select(i, ...) == 1 then
      diceLock[i][2] = 1
    else
      diceLock[i][2] = 0
    end
  end
  frame:updateDiceDisplay()
  frame:sendMyDiceStatus(myGameId)
end

local function sumDice()
  local n = 0
  for i = 1, 5 do
    n = n + diceLock[i][1]
  end
  return n
end

--call chooseField(1..13)
local function chooseField(fieldnumber)
  local _, t = getPlayerTableById(myPlayerId)
  
  if t.field[fieldnumber] ~= -1 then return false end
  
  local tempDice = {0, 0, 0, 0, 0}
  for i = 1, 5 do
    tempDice[i] = diceLock[i][1]
  end
  table.sort(tempDice, function(a,b)return a<b end)
  
  if fieldnumber >= 1 and fieldnumber <= 6 then
    local n = 0
    for i = 1, 5 do
      if tempDice[i] == fieldnumber then
        n = n + fieldnumber
      end
    end
    
    t.field[fieldnumber] = n
    frame:sendMyPlayerTable(myGameId)
    return true
  elseif fieldnumber == 7 then
    local foundThree = false
    local n
    for v = 1, 6 do
      n = 0
      for i = 1, 5 do
        if tempDice[i] == v then
          n = n + 1
        end
      end
      if n >= 3 then
        foundThree = true
        break
      end
    end
    if foundThree then
      t.field[fieldnumber] = sumDice()
      frame:sendMyPlayerTable(myGameId)
      return true
    else
      t.field[fieldnumber] = 0
      frame:sendMyPlayerTable(myGameId)
      return true
    end
  elseif fieldnumber == 8 then
    local foundFour = false
    local n
    for v = 1, 6 do
      n = 0
      for i = 1, 5 do
        if tempDice[i] == v then
          n = n + 1
        end
      end
      if n >= 4 then
        foundFour = true
        break
      end
    end
    if foundFour then
      t.field[fieldnumber] = sumDice()
      frame:sendMyPlayerTable(myGameId)
      return true
    else
      t.field[fieldnumber] = 0
      frame:sendMyPlayerTable(myGameId)
      return true
    end
  elseif fieldnumber == 9 then
    if tempDice[1] == tempDice[2] and tempDice[4] == tempDice[5]
          and (tempDice[2] == tempDice[3] or tempDice[3] == tempDice[4]) then
      --1==2 && 4==5 && (2==3 || 3==4)
      t.field[fieldnumber] = 25
      frame:sendMyPlayerTable(myGameId)
      return true
    else
      t.field[fieldnumber] = 0
      frame:sendMyPlayerTable(myGameId)
      return true
    end
  elseif fieldnumber == 10 or fieldnumber == 11 then
    local streetLength
    if fieldnumber == 10 then
      streetLength = 4
    else
      streetLength = 5
    end
    local foundStreet = false
    for i = 1, 6-streetLength do
      local v = tempDice[i]
      local n = 1
      for v2 = v + 1, v + streetLength - 1 do
        for j = i + 1, 5 do
          if tempDice[j] == v2 then
            n = n + 1
            break
          end
        end
      end
      if n >= streetLength then
        foundStreet = true
        break
      end
    end
    if foundStreet then
      t.field[fieldnumber] = fieldnumber == 10 and 30 or 40
      frame:sendMyPlayerTable(myGameId)
      return true
    else
      t.field[fieldnumber] = 0
      frame:sendMyPlayerTable(myGameId)
      return true
    end
  elseif fieldnumber == 12 then
    local foundKniggel = true
    local v = tempDice[1]
    for i = 2, 5 do
      if tempDice[i] ~= v then
        foundKniggel = false
        break
      end
    end
    if foundKniggel then
      t.field[fieldnumber] = 50
      frame:sendMyPlayerTable(myGameId)
      return true
    else
      t.field[fieldnumber] = 0
      frame:sendMyPlayerTable(myGameId)
      return true
    end
  elseif fieldnumber == 13 then
    t.field[fieldnumber] = sumDice()
    frame:sendMyPlayerTable(myGameId)
    return true
  end
  
  return false
end

local function isChooseAllowed()
  return turnPlayerId == myPlayerId and (myState == 2 or myState == 4 or myState == 6)
end
local function isLockAllowed()
  return turnPlayerId == myPlayerId and (myState == 2 or myState == 4)
end
local function isRollAllowed()
  return turnPlayerId == myPlayerId and (myState == 1 or myState == 3 or myState == 5)
end

function updateGameState()
  logdebug("updateGameState", myState)
  if myState == 1 then
    for i = 1, 5 do
      diceLock[i][1] = 0 --face ?
      diceLock[i][2] = 0 --not locked
    end
    frame:updateDiceDisplay()
    
    if turnPlayerId == myPlayerId then
      print("Kniggel: Your first roll!")
    end
  elseif myState == 2 then
    if turnPlayerId == myPlayerId then
      print("Kniggel: Your first lock or choose!")
    end
  elseif myState == 3 then
    if turnPlayerId == myPlayerId then
      print("Kniggel: Your second roll!")
    end
  elseif myState == 4 then
    if turnPlayerId == myPlayerId then
      print("Kniggel: Your second lock or choose!")
    end
  elseif myState == 5 then
    if turnPlayerId == myPlayerId then
      print("Kniggel: Your third roll!")
    end
  elseif myState == 6 then
    if turnPlayerId == myPlayerId then
      print("Kniggel: Choose now!")
    end
  elseif myState == 7 then
    if myGames[myGameId].invitedBy == myPlayerId then
      local allPlayersFinished = true
      for _, playerId in ipairs(myGames[myGameId].playerlist) do
        local _, t = getPlayerTableById(playerId)
        for i = 1, 13 do
          if t.field[i] == -1 then
            allPlayersFinished = false
            break
          end
        end
      end
      if allPlayersFinished then
        prepareState(8)
      else
        prepareState(1)
      end
    end
  elseif myState == 8 then
    if turnPlayerId == myPlayerId then
      print("Kniggel: You win!")
    end
  end
  
  if isRollAllowed() then
    frame.rollDisplay:SetTextColor(0.67, 1, 0.67, 1)
  else
    frame.rollDisplay:SetTextColor(0.5, 0.5, 0.5, 1)
  end
  if isLockAllowed() then
    frame.diceDisplay[6]:SetTextColor(0.67, 1, 0.67, 1)
  else
    frame.diceDisplay[6]:SetTextColor(0.5, 0.5, 0.5, 1)
  end
  
  if myState == -1 or myState == 8 then
    frame.inviteTextbox:Show()
    frame.inviteDisplay:ShowButton()
    --frame.inviteDisplay.clickFrame:Show()
  else
    frame.inviteTextbox:Hide()
    frame.inviteDisplay:HideButton()
    --frame.inviteDisplay.clickFrame:Hide()
  end
  
  if myState == 0 then
    frame.startGameDisplay:ShowButton()
    --frame.startGameDisplay.clickFrame:Show()
  else
    frame.startGameDisplay:HideButton()
    --frame.startGameDisplay.clickFrame:Hide()
  end
end

local function startGame()
  myState = 0
  
  --remove players who did not respond
  myGames[myGameId].playerlist = {};
  for i, v in ipairs(myGames[myGameId].respondedlist) do
    table.insert(myGames[myGameId].playerlist, v)
    logdebug("Player", v)
  end
  
  nextTurnPlayer = 0
  prepareState(1)
end

--[[
local function getPlayerPointsTop(playerTable)
  local sum = 0
  for i = 1, 6 do
    sum = sum + playerTable[i]
  end
  return sum
end
local function getPlayerPointsBottom(playerTable)
  local sum = 0
  for i = 7, 13 do
    sum = sum + playerTable[i]
  end
  return sum
end
--]]

function updatePlayerColumnContent(playerTable)
  --local text = playerTable.nameshort.."\124n"
  playerTable.collink[1]:SetText(playerTable.nameshort)
  --text = text.."-----\124n"
  local p, sum, topSum, bottomSum = 0, 0, 0, 0
  local f = playerTable.field
  
  for i = 1, 6 do
    p = f[i]
    --text = text + p + "\124n"
    playerTable.collink[i+1]:SetText(p == -1 and "-" or p)
    p = p == -1 and 0 or p
    topSum = topSum + p
  end
  --text = text.."-----\124n"
  
  --text = text..topSum.."\124n"
  playerTable.collink[8]:SetText(topSum)
  if topSum >= 63 then
    topSum = topSum + 35
    --text = text.."35\124n"
    playerTable.collink[9]:SetText("35")
  else
    --text = text.."0\124n"
    playerTable.collink[9]:SetText("0")
  end
  --text = text..topSum.."\124n"
  playerTable.collink[10]:SetText(topSum)
  --text = text.."-----\124n"
  
  for i = 7, 13 do
    p = f[i]
    --text = text + p + "\124n"
    playerTable.collink[i+4]:SetText(p == -1 and "-" or p)
    p = p == -1 and 0 or p
    bottomSum = bottomSum + p
  end
  --text = text.."-----\124n"
  --text = text..bottomSum.."\124n"
  playerTable.collink[18]:SetText(bottomSum)
  --text = text..topSum.."\124n"
  playerTable.collink[19]:SetText(topSum)
  
  sum = topSum + bottomSum
  
  --text = text..sum.."\124n"
  playerTable.collink[20]:SetText(sum)
  
  --playerTable.collink:SetText(text)
  
  return sum
end

local function getNewClickableRowFrame(frameself, tag, tag2)
  local ret = CreateFrame("Frame", nil, frameself)
  ret:SetFrameStrata("LOW")
  ret:SetSize(PLAYER_WIDTH, PLAYER_HEIGHT)
  
  ret.tag = tag
  ret.tag2 = tag2
  
  --HIGHLIGHT
  ret.highlighttexture = ret:CreateTexture(nil, "HIGHLIGHT")
  ret.highlighttexture:SetAllPoints(ret)
  ret.highlighttexture:SetTexture(1, 1, 1, 0.2)
  
  ret.frameRef = frameself
  ret:EnableMouse(true)
  ret:SetScript("OnMouseUp", function(self, button)
    self.frameRef:handleMouseUp(self.tag, self.tag2, button)
  end)
  
  return ret
end

local function getNewClickableFontString(frameself, width, height, anchorself, parent, anchorparent, dx, dy, clickTag1, clickTag2)
  local ret = frameself:CreateFontString(nil, "OVERLAY", "GameFontNormal")
  ret:SetPoint(anchorself, parent, anchorparent, dx, dy)
  ret:SetJustifyH("CENTER")
  ret:SetJustifyV("MIDDLE")
  ret:SetSize(width, height)
  ret:SetTextColor(1, 1, 1, 1)
  
  local froll = getNewClickableRowFrame(frameself, clickTag1, clickTag2)
  froll:SetPoint("TOPLEFT", ret, "TOPLEFT")
  froll:SetSize(width, height)
  ret.clickFrame = froll
  
  ret.ShowButton = function(self)
    self:Show()
    self.clickFrame:Show()
  end
  ret.HideButton = function(self)
    self:Hide()
    self.clickFrame:Hide()
  end
  
  return ret
end

local function initFrame()
  frame:SetPoint(O["framePoint"], O["frameRelativeTo"], O["frameRelativePoint"], O["frameOffsetX"], O["frameOffsetY"])
  frame:SetFrameStrata("LOW")
  frame:SetSize(300, 300)
  
  frame.bgtexture = frame:CreateTexture(nil, "OVERLAY")
  frame.bgtexture:SetAllPoints(frame)
  frame.bgtexture:SetTexture(0, 0, 0, 1)
  
  frame:SetScript("OnDragStart", function(self) self:StartMoving(); end);
  frame:SetScript("OnDragStop", function(self)
    self:StopMovingOrSizing()
    local point, relativeTo, relativePoint, xOfs, yOfs = self:GetPoint()
    O["framePoint"] = point or "LEFT"
    O["frameRelativeTo"] = relativeTo or "UIParent"
    O["frameRelativePoint"] = relativePoint or "CENTER"
    O["frameOffsetX"] = xOfs
    O["frameOffsetY"] = yOfs
  end);
  
  frame.toggleFrame = function(b)
    frame:SetAlpha(b)
    frame:EnableMouse(b==1)
    O.showFrame = b
  end
  
  frame:EnableMouse(true)
  --[[
  frame:SetScript("OnMouseUp", function(self, button)
    self:handleMouseUp(0, button)
  end)
  --]]
  
  frame.tslU = 0
  frame.timerInterval = 10
  
  --------------------
  -- Player table displays
  --------------------
  --[[
  frame.text = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
  frame.text:SetPoint("RIGHT")
  frame.text:SetText("L")
  frame.text:SetTextColor(1, 1, 1, 1)
  --]]
  
  frame.unusedPlayerColumns = {};
  frame.playerColumns = {};
  frame.getNewColumn = function(self)
    local colTable
    if #(self.unusedPlayerColumns) > 0 then
      colTable = table.remove(self.unusedPlayerColumns) --pop last element
      for i, v in ipairs(colTable) do
        v:Show()
      end
    else
      colTable = {};
      local lastRow = nil
      for i = 1, ROW_COUNT do
        local row = self:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        if lastRow then
          row:SetPoint("TOPLEFT", lastRow, "BOTTOMLEFT", 0, 0)
        else
          row:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
        end
        lastRow = row
        row:SetSize(PLAYER_WIDTH, PLAYER_HEIGHT)
        row:SetText("0")
        row:SetJustifyH("LEFT")
        if i >= 2 and i <= 7 or i >= 11 and i <= 17 then
          row:SetTextColor(1, 1, 1, 1)
        else
          row:SetTextColor(0.6, 0.6, 0.8, 1)
        end
        colTable[i] = row
      end
    end
    
    table.insert(self.playerColumns, colTable)
    return colTable
  end
  frame.handleMouseUp = function(self, tag, tag2, button)
    if button == "LeftButton" and not isInEncounter then
      if tag2 == 1 then --choose
        if isChooseAllowed() then
          chooseField(tag)
        end
      elseif tag2 == 2 then
        if isLockAllowed() then
          if tag <= 5 then --lock/unlock
            diceLock[tag][2] = diceLock[tag][2] == 1 and 0 or 1
            self:updateDiceDisplay()
          else --confirm lock
            lockDice(diceLock[1][2], diceLock[2][2], diceLock[3][2], diceLock[4][2], diceLock[5][2])
          end
        end
      elseif tag2 == 3 then --roll
        if isRollAllowed() then
          rollDice()
        end
      elseif tag2 == 4 then --single invite
        inviteNewGame(O.invitelist)
      elseif tag2 == 5 then --start game
        startGame()
      end
    end
  end
  
  frame.addFirstColumn = function(self)
    local colTable = self:getNewColumn()
    
    colTable[1].id = "first"
    
    colTable[1]:SetText("Name:")
    colTable[2]:SetText("1:")
    colTable[3]:SetText("2:")
    colTable[4]:SetText("3:")
    colTable[5]:SetText("4:")
    colTable[6]:SetText("5:")
    colTable[7]:SetText("6:")
    colTable[8]:SetText("SUM:")
    colTable[9]:SetText("Bonus:")
    colTable[10]:SetText("Upper Total:")
    colTable[11]:SetText("3 of a kind:")
    colTable[12]:SetText("4 of a kind:")
    colTable[13]:SetText("Full House:")
    colTable[14]:SetText("Small Street:")
    colTable[15]:SetText("Large Street:")
    colTable[16]:SetText("Kniggel:")
    colTable[17]:SetText("Chance:")
    colTable[18]:SetText("SUM:")
    colTable[19]:SetText("Upper Total:")
    colTable[20]:SetText("TOTAL:")
    
    for i = 1, ROW_COUNT do
      colTable[i]:SetJustifyH("RIGHT")
    end
    
    self.firstColumn = colTable
    
    for i = 1, 6 do
      local f = getNewClickableRowFrame(frame, i, 1)
      f:SetPoint("TOPLEFT", colTable[i + 1], "TOPLEFT")
      colTable[i + 1].clickFrame = f
    end
    for i = 7, 13 do
      local f = getNewClickableRowFrame(frame, i, 1)
      f:SetPoint("TOPLEFT", colTable[i + 4], "TOPLEFT")
      colTable[i + 4].clickFrame = f
    end
  end
  frame.addPlayerColumn = function(self, playerTable)
    local colTable = self:getNewColumn()
    
    colTable[1].id = playerTable.id
    
    playerTable.collink = colTable
  end
  frame.removePlayerColumn = function(self, playerTable)
    --[[
    for i, col in ipairs(self.playerColumns) do
      if col.name == playerTable.name then
        local col2 = table.remove(self.playerColumns, i) --pop element
        col2:Hide()
        playerTable.collink = nil
        table.insert(self.unusedPlayerColumns, col2)
        return
      end
    end
    --]]
    for i, colTable in ipairs(self.playerColumns) do
      if colTable[1].id == playerTable.id then
        local colTable2 = table.remove(self.playerColumns, i) --pop element
        for _, row in ipairs(colTable2) do
          row:Hide()
        end
        playerTable.collink = nil
        table.insert(self.unusedPlayerColumns, colTable2)
        return
      end
    end
  end
  
  frame.alignPlayerColumns = function(self)
    --[[
    for i, col in ipairs(self.playerColumns) do
      col:SetPoint("TOPLEFT", self, "TOPLEFT", (i+1)*PLAYER_WIDTH, 0)
    end
    --]]
    for i, colTable in ipairs(self.playerColumns) do
      colTable[1]:SetPoint("TOPLEFT", self, "TOPLEFT", (i-1)*(PLAYER_WIDTH+COL_SPACING), 0)
    end
    self:SetWidth((PLAYER_WIDTH + COL_SPACING) * (#(self.playerColumns)))
    
    for i = 1, 6 do
      self.firstColumn[i + 1].clickFrame:SetWidth(self:GetWidth())
    end
    for i = 7, 13 do
      self.firstColumn[i + 4].clickFrame:SetWidth(self:GetWidth())
    end
  end
  
  --[[
  frame.gamefield = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
  frame.gamefield:SetPoint("TOPLEFT", frame, "TOPLEFT")
  frame.gamefield:SetSize(PLAYER_WIDTH * 2)
  frame.gamefield:SetText("L")
  frame.gamefield:SetTextColor(1, 1, 1, 1)
  --]]
  
  frame.OnUpdate = function(self, elapsed)
    self.tslU = self.tslU + elapsed
    if (self.tslU >= self.timerInterval) then
      self.tslU = 0
      --
    end
  end
  frame:SetScript("OnUpdate", function(self, elapsed)
    self.OnUpdate(self, elapsed)
  end);
  
  frame:addFirstColumn()
  
  --------------------
  -- Dice display
  --------------------
  
  frame.diceDisplay = {};
  local lastDice = nil
  for i = 1, 6 do
    local t = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    frame.diceDisplay[i] = t
    
    if lastDice then
      t:SetPoint("BOTTOMLEFT", lastDice, "BOTTOMRIGHT")
    else
      t:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT")
    end
    lastDice = t
    if i <= 5 then
      t:SetText("d"..i)
    else
      t:SetText("OK")
    end
    t:SetJustifyH("CENTER")
    t:SetJustifyV("MIDDLE")
    t:SetSize(DICE_WIDTH, DICE_WIDTH)
    t:SetTextColor(1, 1, 1, 1)
    
    local f = getNewClickableRowFrame(frame, i, 2)
    f:SetPoint("TOPLEFT", t, "TOPLEFT")
    f:SetSize(DICE_WIDTH, DICE_WIDTH)
    t.clickFrame = f
  end
  frame.updateDiceDisplay = function(self)
    for i = 1, 5 do
      local text
      if diceLock[i][2] == 1 then
        text = "["..diceLock[i][1].."]"
      else
        text = diceLock[i][1]
      end
      frame.diceDisplay[i]:SetText(text)
    end
  end
  
  --------------------
  -- Roll display
  --------------------
  
  frame.rollDisplay = getNewClickableFontString(frame, ROLL_WIDTH, DICE_WIDTH, "BOTTOMLEFT", lastDice, "BOTTOMRIGHT", 10, 0, 1, 3)
  frame.rollDisplay:SetText("ROLL")
  
  --------------------
  -- Invite display
  --------------------
  
  local invtextbox = CreateFrame("EditBox", nil, frame, "InputBoxTemplate")
  frame.inviteTextbox = invtextbox
  invtextbox:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 10, DICE_WIDTH)
  invtextbox:SetAutoFocus(false)
  invtextbox:SetWidth(PLAYER_WIDTH)
  invtextbox:SetHeight(DICE_WIDTH)
  invtextbox:SetTextInsets(0,0,3,3)
  --invtextbox:SetNumeric()
  invtextbox:SetText(O.invitelist)
  invtextbox:SetCursorPosition(0)
  invtextbox:SetScript("OnEnterPressed", function(self)
    self:ClearFocus()
    O.invitelist = self:GetText()
  end);
  invtextbox:SetScript("OnEscapePressed", function(self)
    self:ClearFocus()
    self:SetText(O.invitelist)
  end);
  
  frame.inviteDisplay = getNewClickableFontString(frame, ROLL_WIDTH, DICE_WIDTH, "BOTTOMLEFT", invtextbox, "BOTTOMRIGHT", 10, 0, 1, 4)
  frame.inviteDisplay:SetText("Invite")
  
  frame.startGameDisplay = getNewClickableFontString(frame, ROLL_WIDTH, DICE_WIDTH, "BOTTOMLEFT", frame.inviteDisplay, "BOTTOMRIGHT", 10, 0, 1, 5)
  frame.startGameDisplay:SetText("Start")
  
  --------------------
  -- COM stuff
  --------------------
  frame.COMprefix = "KNiGGEL"
  frame.COMchannel = "GUILD"--"RAID"
  frame.COMPlayerName = UnitName("player")
  frame.COMRealmName = GetRealmName("player")
  myPlayerId = frame.COMPlayerName.."-"..frame.COMRealmName
  frame.COMMsgPrefixes = {
    sendOwnPlayerTable = "PTA";
    sendOwnDiceStatus = "PDS";
    sendBroadcastInvite = "IVB";
    sendPersonalInvite = "IVB";
    sendInviteAccept = "IVA";
    sendBroadcastNewPlayerJoined = "PNW";
    sendBroadcastPlayersTurn = "STA";
    sendBroadcastPlayerWins = "END";
  };
  --frame.COMlastStatusRequestSent = -1
  local COMok = RegisterAddonMessagePrefix(frame.COMprefix)
  
  frame.getNewGameId = function(self)
    return myPlayerId..time()
  end
  frame.sendMyPlayerTable = function(self, gameId)
    --local _, t = getPlayerTableById(self.COMPlayerName, self.COMRealmName)
    local _, t = getPlayerTableById(myPlayerId)
    if t then
      local msg = self.COMMsgPrefixes.sendOwnPlayerTable
      msg = msg..gameId
      for i = 1, 13 do
        msg = msg.." "..t["field"][i]
      end
      --msgcontent = "Playername-Realmname1118722038 <field1 value>[...] <field13 value>"
      iTP:SendAddonMessage(self.COMprefix, msg, self.COMchannel)
    end
  end
  frame.sendMyDiceStatus = function(self, gameId)
    local msg = self.COMMsgPrefixes.sendOwnDiceStatus
    msg = msg..gameId
    for i = 1, 5 do
      msg = msg.." "..diceLock[i][1].." "..diceLock[i][2] --dice value, dice lock status
    end
    --msgcontent = "Playername-Realmname1118722038 <dice1 value> <dice1 locked>[...] <dice5 value> <dice5 locked>"
    iTP:SendAddonMessage(self.COMprefix, msg, self.COMchannel)
  end
  frame.sendBroadcastInvite = function(self, gameId)
    local msg = self.COMMsgPrefixes.sendBroadcastInvite
    msg = msg..gameId
    --msgcontent = "Playername-Realmname1118722038"
    iTP:SendAddonMessage(self.COMprefix, msg, self.COMchannel)
  end
  frame.sendPersonalInvite = function(self, gameId, playerId)
    local msg = self.COMMsgPrefixes.sendPersonalInvite
    msg = msg..gameId
    --msgcontent = "Playername-Realmname1118722038"
    iTP:SendAddonMessage(self.COMprefix, msg, "WHISPER", playerId)
  end
  frame.sendInviteAccept = function(self, gameId, playerId)
    local msg = self.COMMsgPrefixes.sendInviteAccept
    msg = msg..gameId
    --msgcontent = "Playername-Realmname1118722038"
    iTP:SendAddonMessage(self.COMprefix, msg, "WHISPER", playerId)
  end
  frame.sendBroadcastNewPlayerJoined = function(self, gameId, playerId)
    local msg = self.COMMsgPrefixes.sendBroadcastNewPlayerJoined
    msg = msg..gameId.." "..playerId
    --msgcontent = "Playername-Realmname1118722038 NewPlayername-Realmname"
    iTP:SendAddonMessage(self.COMprefix, msg, self.COMchannel)
  end
  frame.sendBroadcastPlayersTurn = function(self, gameId, state, playerId)
    local msg = self.COMMsgPrefixes.sendBroadcastPlayersTurn
    msg = msg..gameId.." "..state.." "..playerId
    --msgcontent = "Playername-Realmname1118722038 1 Playername-Realmname"
    iTP:SendAddonMessage(self.COMprefix, msg, self.COMchannel)
  end
  frame.sendBroadcastPlayerWins = function(self, gameId, state, playerId)
    local msg = self.COMMsgPrefixes.sendBroadcastPlayerWins
    msg = msg..gameId.." "..state.." "..playerId
    --msgcontent = "Playername-Realmname1118722038 8 Playername-Realmname"
    iTP:SendAddonMessage(self.COMprefix, msg, self.COMchannel)
  end
  
  --------------------
  -- Popup frame
  --------------------
  
  popupFrame:SetPoint(O.framePointPopup, O.frameRelativeToPopup, O.frameRelativePointPopup, O.frameOffsetXPopup, O.frameOffsetYPopup)
  popupFrame:SetFrameStrata("LOW")
  popupFrame:SetSize(166, 70)
  
  popupFrame.bgtexture = popupFrame:CreateTexture(nil, "OVERLAY")
  popupFrame.bgtexture:SetAllPoints(popupFrame)
  popupFrame.bgtexture:SetTexture(0, 0, 0, 1)
  
  local qstring = popupFrame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
  qstring:SetPoint("TOP", popupFrame, "TOP", 0, -2)
  qstring:SetTextColor(1, 1, 1, 1)
  qstring:SetText("Accept invite from:")
  popupFrame.qstring = qstring
  
  local qstring2 = popupFrame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
  qstring2:SetPoint("TOP", qstring, "BOTTOM", 0, -2)
  qstring2:SetTextColor(0.5, 0.5, 1, 1)
  qstring2:SetText("Playername-Realmname")
  popupFrame.qstring2 = qstring2
  
  popupFrame.acceptButton = getNewClickableFontString(popupFrame, 80, 20, "BOTTOMRIGHT", popupFrame, "BOTTOM", -1, 0, 1, 1)
  popupFrame.acceptButton:SetText("Accept")
  popupFrame.declineButton = getNewClickableFontString(popupFrame, 80, 20, "BOTTOMLEFT", popupFrame, "BOTTOM", 1, 0, 1, 2)
  popupFrame.declineButton:SetText("Decline")
  
  popupFrame.frameRef = frame
  
  popupFrame.handleMouseUp = function(self, tag, tag2, button)
    if button == "LeftButton" then
      if tag2 == 1 then --accept
        self.frameRef:sendInviteAccept(invitedByGameId, invitedBy)
        myGameId = invitedByGameId
        buildGameField(myGameId)
        self:Hide()
      elseif tag2 == 2 then --decline
        self:Hide()
      end
    end
  end
  
  popupFrame:Hide()
  popupFrame.showInvite = function(self, from)
    popupFrame.qstring2:SetText(from)
    popupFrame:Show()
  end
end

function frameEvents:PLAYER_ENTERING_WORLD(...)
  frame:UnregisterEvent("PLAYER_ENTERING_WORLD")
  
  --------------------
  -- get all options/upgrade option table
  --------------------
  if not KniggelOptions then
    KniggelOptions = DefaultO
  end
  O = KniggelOptions
  
  O.addonVersion = GetAddOnMetadata("LookingForSatchels", "Version")
  
  O.showFrame = O.showFrame or DefaultO.showFrame
  O.invitelist = O.invitelist or DefaultO.invitelist
  O.fixed = O.fixed or DefaultO.fixed
  
  O.framePointPopup = O.framePointPopup or DefaultO.framePointPopup
  O.frameRelativeToPopup = O.frameRelativeToPopup or DefaultO.frameRelativeToPopup
  O.frameRelativePointPopup = O.frameRelativePointPopup or DefaultO.frameRelativePointPopup
  O.frameOffsetXPopup = O.frameOffsetXPopup or DefaultO.frameOffsetXPopup
  O.frameOffsetYPopup = O.frameOffsetYPopup or DefaultO.frameOffsetYPopup
  
  --------------------
  -- init frame and frame's functions
  --------------------
  initFrame()
  
  frame.toggleFrame(O.showFrame)
  if O.fixed == 0 then
    frame:SetMovable(true)
    frame:RegisterForDrag("LeftButton")
  end
  
  updateGameState()
  
  frame:RegisterEvent("PLAYER_REGEN_ENABLED")
  frame:RegisterEvent("PLAYER_REGEN_DISABLED")
  frame:RegisterEvent("ENCOUNTER_END")
  frame:RegisterEvent("ENCOUNTER_START")
  --frame:RegisterEvent("CHAT_MSG_ADDON")
end

--function frameEvents:CHAT_MSG_ADDON(prefix, msg, channel, from)
function iTPCallback:CHAT_MSG_ADDON(prefix, msg, channel, from, sendermsgid)
  if prefix == frame.COMprefix then
    logdebug(prefix, ";", msg, ";", channel, ";", from)
    local MSGPrefix = msg:sub(1, 3)
    if myState == 0 then --only listen for invites/accepts
      if MSGPrefix == frame.COMMsgPrefixes.sendPersonalInvite and channel == "WHISPER" then
        print("Kniggel invite from:", from)
        invitedBy = from
        invitedByGameId = msg:sub(4)
        addGame(invitedByGameId, invitedBy)
        addPlayerToGame(invitedByGameId, invitedBy)
        popupFrame:showInvite(from)
      --[[
      elseif MSGPrefix == frame.COMMsgPrefixes.sendBroadcastInvite then
        print("Kniggel invite from:", from)
        invitedBy = from
        invitedByGameId = msg:sub(4)
      --]]
      elseif MSGPrefix == frame.COMMsgPrefixes.sendInviteAccept and channel == "WHISPER" and msg:sub(4) == myGameId then
        if isPlayerInvited(myGameId, from) then
          addPlayerTable(from)
          sendBroadcastNewPlayerJoined(myGameId, from)
          addPlayerToGameResponded(myGameId, from)
        end
      elseif MSGPrefix == frame.COMMsgPrefixes.sendBroadcastNewPlayerJoined then
        local gameId, playerId = unpack(mysplit2(msg:sub(4)))
        
        if gameId and playerId then
          addPlayerToGame(gameId, playerId)
          if gameId == myGameId then
            addPlayerTable(playerId)
          end
        end
      elseif MSGPrefix == frame.COMMsgPrefixes.sendBroadcastPlayersTurn then
        local gameId, state, playerId = unpack(mysplit2(msg:sub(4)))
        if gameId and state and playerId then
          if gameId == myGameId then
            myState = tonumber(state)
            turnPlayerId = playerId
            updateGameState()
          end
        end
      end
    else
      if isPlayerInvited(myGameId, from) then
        if MSGPrefix == frame.COMMsgPrefixes.sendOwnPlayerTable then
          local playerTableFields = mysplit2(msg:sub(4))
          --TODO: assign player table, update display
          if (not playerTableFields) or (#playerTableFields ~= 14) then
            error(frame.COMMsgPrefixes.sendOwnPlayerTable.." corrupted data received")
          end
          if playerTableFields[1] == myGameId then
            local _, v = getPlayerTableById(from)
            for i = 1, 13 do
              v.field[i] = tonumber(playerTableFields[i + 1])
            end
            updatePlayerColumnContent(v)
            if myGames[myGameId].invitedBy == myPlayerId then
              if myState == 2 then
                prepareState(7)
              elseif myState == 4 then
                prepareState(7)
              elseif myState == 6 then
                prepareState(7)
              end
            end
          end
        elseif MSGPrefix == frame.COMMsgPrefixes.sendOwnDiceStatus then
          local diceTable = mysplit2(msg:sub(4))
          if (not diceTable) or (#diceTable ~= 11) then
            error(frame.COMMsgPrefixes.sendOwnDiceStatus.." corrupted data received")
          end
          if diceTable[1] == myGameId then
            logdebug("Got the dice from player", from)
            for i = 1, 5 do
              diceLock[i][1] = tonumber(diceTable[i * 2]) --value
              diceLock[i][2] = tonumber(diceTable[i * 2 + 1]) --locked status
            end
            frame:updateDiceDisplay()
            if myGames[myGameId].invitedBy == myPlayerId then
              if myState == 1 then --player rolled
                prepareState(2)
              elseif myState == 2 then --player locked
                prepareState(3)
              elseif myState == 3 then --player rolled
                prepareState(4)
              elseif myState == 4 then --player locked
                prepareState(5)
              elseif myState == 5 then --player rolled
                prepareState(6)
              end
            end
          end
        elseif MSGPrefix == frame.COMMsgPrefixes.sendBroadcastPlayersTurn then
          local gameId, state, playerId = unpack(mysplit2(msg:sub(4)))
          if gameId and state and playerId then
            if gameId == myGameId then
              myState = tonumber(state)
              turnPlayerId = playerId
              updateGameState()
            end
          end
        elseif MSGPrefix == frame.COMMsgPrefixes.sendBroadcastPlayerWins then
          local gameId, state, playerId = unpack(mysplit2(msg:sub(4)))
          if gameId and state and playerId then
            if gameId == myGameId then
              myState = tonumber(state)
              turnPlayerId = playerId
              updateGameState()
            end
          end
        end
      else
        --logdebug("player was not invited!", from)
      end
    end
  end
end
function frameEvents:PLAYER_REGEN_ENABLED()
end
function frameEvents:PLAYER_REGEN_DISABLED()
end
function frameEvents:ENCOUNTER_END()
  isInEncounter = false
end
function frameEvents:ENCOUNTER_START()
  isInEncounter = true
end

frame:SetScript("OnEvent", function(self, event, ...)
  frameEvents[event](self, ...)
end);
frame:RegisterEvent("PLAYER_ENTERING_WORLD")

function iTPCallback:OnSendAddonMessageBegin(msgid, partcount)
  --logdebug("OnSendAddonMessageBegin", msgid, partcount)
end
function iTPCallback:OnSendAddonMessageProgress(msgid, partcount, partnumber)
  --logdebug("OnSendAddonMessageProgress", msgid, partcount, partnumber)
end
function iTPCallback:OnSendAddonMessageEnd(msgid, partcount)
  --logdebug("OnSendAddonMessageEnd", msgid, partcount)
end

--split string containing quoted and non quoted arguments
--input pattern: (\S+|".+")?(\s+(\S+|".+"))*
--example input: [[arg1 "arg2part1 arg2part2" arg3]]
--example output: {"arg1", "arg2part1 arg2part2", "arg3"}
function mysplit2(inputstr, separator)
  local i, i1, i2, l, ret, retI = 1, 0, 0, inputstr:len(), {}, 1
  if not separator then
    separator = "%s"
  end
  --remove leading spaces
  i1, i2 = inputstr:find("^"..separator.."+")
  if i1 then
    i = i2 + 1
  end
  
  while i <= l do
    --find end of current arg
    if (inputstr:sub(i, i)) == "\"" then
      --quoted arg, find end quote
      i1, i2 = inputstr:find("\""..separator.."+", i + 1)
      if i1 then
        --spaces after end quote, more args to follow
        ret[retI] = inputstr:sub(i + 1, i1 - 1)
        retI = retI + 1
        i = i2 + 1
      else
        i1, i2 = inputstr:find("\"$", i + 1)
        if i1 then
          --end of msg
          ret[retI] = inputstr:sub(i + 1, i1 - 1)
          return ret
        else
          -- no end quote found, or end quote followed by no-space-charater found, disregard last arg
          return ret
        end
      end
    else
      --not quoted arg, find next space (if any)
      i1, i2 = inputstr:find(separator.."+", i + 1)
      if i1 then
        --spaces after arg, more args to follow
        ret[retI] = inputstr:sub(i, i1 - 1)
        retI = retI + 1
        i = i2 + 1
      else
        --end of msg
        ret[retI] = inputstr:sub(i)
        return ret
      end
    end
  end
  
  return ret
end

SLASH_KNIGGEL1 = "/kniggel"
SlashCmdList["KNIGGEL"] = function(msg, editbox)
  args = mysplit2(msg or "")
  local arg1 = string.lower(args[1] or "")
  if arg1 == "move" then
    if O.fixed == 0 then
      O.fixed = 1
      frame:SetMovable(false)
      frame:RegisterForDrag("")
    else
      O.fixed = 0
      frame:SetMovable(true)
      frame:RegisterForDrag("LeftButton")
    end
    print("|cffaaaaffKniggel |rmove |cffaaaaffis now "..(O.fixed == 0 and "|cffaaffaamoving" or "|cffff8888fixed"))
  elseif arg1 == "hide" then
    frame.toggleFrame(0)
    print("|cffaaaaffKniggel |rhidden")
  elseif arg1 == "show" then
    frame.toggleFrame(1)
    print("|cffaaaaffKniggel |rshown")
  elseif arg1 == "reset" then
    O["framePoint"] = DefaultO["framePoint"]
    O["frameRelativeTo"] = DefaultO["frameRelativeTo"]
    O["frameRelativePoint"] = DefaultO["frameRelativePoint"]
    O["frameOffsetX"] = DefaultO["frameOffsetX"]
    O["frameOffsetY"] = DefaultO["frameOffsetY"]
    
    frame:ClearAllPoints()
    frame:SetPoint(O["framePoint"], O["frameRelativeTo"], O["frameRelativePoint"], O["frameOffsetX"], O["frameOffsetY"])
    
    print("|cffaaaaffKniggel |rposition reset")
  elseif arg1 == "addplayer" then
    addPlayerTable(args[2].."-"..args[3]) --playername-realm
  elseif arg1 == "testaddandsendself" then
    addPlayerTable(myPlayerId)
    frame:sendMyPlayerTable()
  elseif arg1 == "testsenddice" then
    frame:sendMyDiceStatus(myGameId)
  elseif arg1 == "testbroadcastinvite" then
    frame:sendBroadcastInvite()
  elseif arg1 == "testsingleinvites" then
    inviteNewGame(args[2])
  elseif arg1 == "testinviteaccept" then
    frame:sendInviteAccept(invitedByGameId, invitedBy)
    myGameId = invitedByGameId
    buildGameField(myGameId)
  elseif arg1 == "teststartgame" then
    startGame()
  elseif arg1 == "testroll" then
    if turnPlayerId == myPlayerId and (myState == 1 or myState == 3 or myState == 5) then
      rollDice()
    end
  elseif arg1 == "testlock" then
    if turnPlayerId == myPlayerId and (myState == 2 or myState == 4) then
      lockDice(tonumber(args[2]), tonumber(args[3]), tonumber(args[4]), tonumber(args[5]), tonumber(args[6]))
    end
  elseif arg1 == "testchoose" then
    if turnPlayerId == myPlayerId and (myState == 2 or myState == 4 or myState == 6) then
      chooseField(tonumber(args[2]))
    end
  elseif arg1 == "channel" then
    frame.COMchannel = args[2]
    logdebug("COM channel changed to:", frame.COMchannel)
  elseif arg1 == "debug" then
    isdebug = not isdebug
    iTP:toggledebug(isdebug)
    print("isdebug", isdebug)
    --[[
    /kniggel channel PARTY
    /kniggel testsingleinvites Ingenium2-Blackmoore
    /kniggel testinviteaccept
    /kniggel teststartgame
    /kniggel testroll
    /run KNIGGELDICELOCK[1][1]=5;KNIGGELDICELOCK[2][1]=3;KNIGGELDICELOCK[3][1]=4;KNIGGELDICELOCK[4][1]=3;KNIGGELDICELOCK[5][1]=2;
    /kniggel testlock 1 0 1 0 0
    /kniggel testchoose 10
    --]]
  else
    print("|cffaaaaffKniggel |r"..(GetAddOnMetadata("Kniggel", "Version") or "").." |cffaaaaff(use |r/kniggel <option> |cffaaaafffor these options)")
    print("  move |cffaaaafftoggle moving the frame ("..(O.fixed == 0 and "|cffaaffaamoving" or "|cffff8888fixed").."|cffaaaaff)")
    print("  reset |cffaaaaffreset the frame's position")
    print("  show/hide |cffaaaaffshow/hide the indicator ("..(O.showFrame == 1 and "|cffaaffaashown" or "|cffff8888hidden").."|cffaaaaff)")
  end
end
