--2016 by Inge (CC BY-SA 3.0)
local iTP = LibStub:NewLibrary("iTransferProtocol-1.0", 1)

if not iTP then return end -- No upgrade needed

local iTPPrefix = "iTP1"

iTP.prefixes = {};
local prefixesList = {};
local nextPrefix = 1
local lastSentTimestamp = 0
local isdebug = false
local function logdebug(...)
  if isdebug then
    print(...)
  end
end

local iTPFrame = CreateFrame("Frame", nil, UIParent)
iTPFrame.tslu = 0
iTPFrame.interval = 0.5

local function trim1(s)
  return (s:gsub("^%s*(.-)%s*$", "%1"))
end
local function getPaddedPrefix(prefix)
  return ("%s        "):format(prefix):sub(1, 8) --only get the first 8 chars, padded with spaces
end

function iTP:toggledebug(b)
  isdebug = b
  print("|cffffff00Debug:", isdebug)
end

function iTP:RegisterPrefix(prefix)
  prefix = getPaddedPrefix(prefix)
  if self.prefixes[prefix] then
    error(format("Prefix %s is already registered.", trim1(prefix)))
  end
  
  self.prefixes[prefix] = {
    ["prefix"] = prefix;
    ["nextmsgid"] = 1; --"0001"
    ["messagesOut"] = {};
    ["messagesIn"] = {};
  };
  table.insert(prefixesList, self.prefixes[prefix])
  
  return self.prefixes[prefix]
end

function iTP:SendAddonMessage(prefix, message, channel, whispertarget)
  prefix = getPaddedPrefix(prefix)
  prefixTable = self.prefixes[prefix]
  if not prefixTable then
    error(format("Prefix %s is not registered.", trim1(prefix)))
  end
  
  if message:len() > 15073050 then --"ffff" parts with len 230 (15,073,050 bytes)
    error("Message too long.")
  end
  
  local parts = {};
  if message:len() == 0 then
    table.insert(parts, "")
  else
    for i = 1, message:len(), 230 do
      table.insert(parts, message:sub(i, i + 229))
    end
  end
  
  local messageOut = {
    ["msgid"] = prefixTable.nextmsgid;
    ["channel"] = channel;
    ["whispertarget"] = whispertarget;
    ["partcount"] = #parts;
    ["parts"] = parts;
    ["nextpartnumber"] = 1;
  };
  table.insert(prefixTable.messagesOut, messageOut)
  
  if prefixTable.OnSendAddonMessageBegin and type(prefixTable.OnSendAddonMessageBegin) == "function" then
    local ok, err = pcall(prefixTable.OnSendAddonMessageBegin, prefixTable, prefixTable.nextmsgid, #parts)
  end
  
  prefixTable.nextmsgid = prefixTable.nextmsgid + 1
  if prefixTable.nextmsgid > 65535 then --"ffff"
    prefixTable.nextmsgid = 1
  end
  
  if lastSentTimestamp < (GetTime() - iTPFrame.interval) then
    iTPFrame:OnUpdate(iTPFrame.interval)
  end
  
  return messageOut.msgid
end

local function SendNextMessage()
  if #(prefixesList) == 0 then return end
  
  local prefixIndex = 0
  
  for i = nextPrefix, #(prefixesList) do
    if #(prefixesList[i].messagesOut) > 0 then
      prefixIndex = i
      break
    end
  end
  if prefixIndex == 0 then
    for i = 1, nextPrefix - 1 do
      if #(prefixesList[i].messagesOut) > 0 then
        prefixIndex = i
        break
      end
    end
  end
  
  if prefixIndex == 0 then return end
  
  local message = prefixesList[prefixIndex].messagesOut[1]
  
  local msg = format("%s%04x%04x%04x%s", prefixesList[prefixIndex].prefix, message.msgid, message.partcount, message.nextpartnumber, message.parts[message.nextpartnumber])

  SendAddonMessage(iTPPrefix, msg, message.channel, message.whispertarget)
  
  if prefixesList[prefixIndex].OnSendAddonMessageProgress and type(prefixesList[prefixIndex].OnSendAddonMessageProgress) == "function" then
    local ok, err = pcall(prefixesList[prefixIndex].OnSendAddonMessageProgress, prefixesList[prefixIndex], message.msgid, message.partcount, message.nextpartnumber)
  end

  message.parts[message.nextpartnumber] = nil
  
  message.nextpartnumber = message.nextpartnumber + 1
  if message.nextpartnumber > message.partcount then
    if prefixesList[prefixIndex].OnSendAddonMessageEnd and type(prefixesList[prefixIndex].OnSendAddonMessageEnd) == "function" then
      local ok, err = pcall(prefixesList[prefixIndex].OnSendAddonMessageEnd, prefixesList[prefixIndex], message.msgid, message.partcount)
    end
    
    table.remove(prefixesList[prefixIndex].messagesOut, 1)
  end
  
  nextPrefix = prefixIndex + 1
  if nextPrefix > #(prefixesList) then
    nextPrefix = 1
  end
  
  lastSentTimestamp = GetTime()
end

function iTPFrame:CHAT_MSG_ADDON(prefix, msg, channel, from)
  if prefix ~= iTPPrefix then return end
  logdebug("|cffffff00CHAT_MSG_ADDON", prefix, msg, channel, from)

  local payloadPrefix = msg:sub(1, 8)
  
  prefixTable = iTP.prefixes[payloadPrefix]
  if not prefixTable then return end
  
  local msgid = tonumber(msg:sub(9, 12), 16)
  local partcount = tonumber(msg:sub(13, 16), 16)
  local partnumber = tonumber(msg:sub(17, 20), 16)
  local payload = msg:sub(21)
  
  if not prefixTable.messagesIn[from] then
    prefixTable.messagesIn[from] = {};
  end
  local playeridTable = prefixTable.messagesIn[from]
  
  if not playeridTable[msgid] then
    playeridTable[msgid] = {
      ["channel"] = channel;
      ["partcount"] = partcount;
      ["parts"] = {};
      ["nextpartnumber"] = 1;
    };
  end
  local messageTable = playeridTable[msgid]
  
  if partnumber ~= messageTable.nextpartnumber then
    logdebug(format("|cffffff00Got part %d, expected %d.", partnumber, messageTable.nextpartnumber))
    playeridTable[msgid] = nil
    return
  end
  
  messageTable.parts[partnumber] = payload
  messageTable.nextpartnumber = messageTable.nextpartnumber + 1
  
  if messageTable.nextpartnumber > messageTable.partcount then
    local msgconcat = table.concat(messageTable.parts, "")
    if prefixTable.CHAT_MSG_ADDON and type(prefixTable.CHAT_MSG_ADDON) == "function" then
      local ok, err = pcall(prefixTable.CHAT_MSG_ADDON, prefixTable, trim1(payloadPrefix), msgconcat, channel, from, msgid)
    end
    
    playeridTable[msgid] = nil
  end
end
iTPFrame:RegisterEvent("CHAT_MSG_ADDON")

iTPFrame:SetScript("OnEvent", function(self, event, ...)
  iTPFrame[event](self, ...)
end);

function iTPFrame:OnUpdate(elapsed)
  self.tslu = self.tslu + elapsed
  if self.tslu >= self.interval then
    SendNextMessage()
    self.tslu = 0
  end
end
iTPFrame:SetScript("OnUpdate", function(self, elapsed)
  self:OnUpdate(elapsed)
end);